{-# LANGUAGE MultiParamTypeClasses
,FlexibleInstances
,TypeSynonymInstances
,FlexibleContexts
,FunctionalDependencies
,ScopedTypeVariables
,KindSignatures #-}

{- A typed dataflow language -}

module Data(Repr(..)
            ,PrimType(..)
            ,MRepr(..)
            ,Data(..)
            ,Field(..)
            ,Ring(..)
            ,TypeOf(..)
            ,Tint,Treal,Tbool
            ,typeVec
            ) where

-- import StateCont
import Control.Monad
import Control.Monad.Identity
import Prelude hiding (div)
import Type

{- Primitive data types -}
type Treal  = Double
type Tint   = Int
type Tbool  = Bool

class PrimType t
instance PrimType Tbool
instance PrimType Tint
instance PrimType Treal


typeVec (Type t d) = Type t (d + 1)
  
     

class TypeOf t where typeOf :: t -> Type

instance TypeOf Tbool where typeOf _ = Type ABool 0
instance TypeOf Tint  where typeOf _ = Type AInt  0
instance TypeOf Treal where typeOf _ = Type AReal 0

class Repr r where
  unRepr :: r t -> t
  unRepr = undefined -- combines with typeOf
  
class (Monad m, Repr r) => MRepr m r | r -> m, m -> r where
  unMRepr :: m (r t) -> r t
  unMRepr = undefined -- combines with typeOf


class MRepr m r => Ring m r t | r -> m where
  {- Move literals elsewhere.  It's best to keep the instance clean:
     only keep operations that all Rings support.
  lit  :: t -> m (r t) -}
  
  zero :: r t
  one  :: r t
  
  add  :: r t -> r t -> m (r t) 
  sub  :: r t -> r t -> m (r t) 
  mul  :: r t -> r t -> m (r t)

  eq   :: r t -> r t -> m (r Tbool)
  lt   :: r t -> r t -> m (r Tbool)  -- Elswhere?


class Ring m r t => Field m r t | r -> m, m -> r where

  div  :: r t -> r t -> m (r t)



class (Ring  m r Tint
      ,Field m r Treal) 
      => Data m r | r -> m, m -> r where
  
  i2f :: r Tint  -> m (r Treal)
  f2i :: r Treal -> m (r Tint)

  litb :: Tbool -> r Tbool
  liti :: Tint  -> r Tint
  litr :: Treal -> r Treal

