{-# LANGUAGE TypeSynonymInstances 
, FlexibleInstances
, FlexibleContexts
, MultiParamTypeClasses #-}

-- See Term.hs Value.hs in dspm/ for earlier version.

module Eval(Val(..), VVal(..), Eval(..), eval, evalSys, evalSig) where

import Control.Monad.Identity
import Prelude hiding (div)
import Data
import Siso
import Vec
           
{- Interpret by evaluating has Haskell code:
   r = Val:  a simple tagged Haskell value
   m = Eval: a tagged Identity monad -}
 
newtype Val  t = Val  { unVal  :: t } deriving (Eq, Show)
newtype Eval v = Eval { unEval :: Identity v }

instance Monad Eval where
  return = Eval . return
  (Eval mid) >>= f = Eval $ (mid >>= unEval . f)

instance Repr Val
instance MRepr Eval Val

op2 f (Val a) (Val b) = return $ Val $ f a b

instance Ring Eval Val Treal where 
  add = op2 (+)
  sub = op2 (-)
  mul = op2 (*)
  eq  = op2 (==)
  lt  = op2 (<)
  -- lit = return . Val
  zero = Val 0
  one  = Val 1

instance Field Eval Val Treal where 
  div = op2 (/)

instance Ring Eval Val Tint where 
  add = op2 (+)
  sub = op2 (-)
  mul = op2 (*)
  
  eq  = op2 (==)
  lt  = op2 (<)
   
  -- lit = return . Val
  zero = Val 0
  one  = Val 1
  
instance Data Eval Val where              
  f2i = undefined
  i2f = undefined
  litb = Val
  liti = Val
  litr = Val


-- Type-directed evaluation of Sys to sequence processor.
evalSys :: Sys Eval (Val i) (Val o) -> [i] -> [o]
evalSys sys seq = map unVal os where
  Eval (Identity os) = (mapSys sys) $ map Val seq

evalSig sig = evalSys sig $ cycle [()]

eval = runIdentity . unEval
  
-- For convenience, evaluate the Eval when printing.
instance Show t => Show (Eval t) where
  show (Eval x) = "(Eval " ++ (show $ runIdentity x) ++ ")"

-- The interface with Sys is a dummy; we can't do any compilation.
instance SysState Eval (Val t) where
  genStateIn  = return
  saveState _ = return ()
  

instance Vec VVal

newtype VVal t = VVal [t]

instance VecFold VVal Eval Val e a where
  vfold f a (Val (VVal [])) = return a
  vfold f a (Val (VVal (x:xs))) = do
    a' <- f (Val x) a
    vfold f a' $ Val $ VVal xs
    
    
